from config import *
import arcade as ar
import numpy as np
import random
from collections import deque
import math


class BodySegment:
    def __init__(self, color, pos, head=False):
        self.pos = pos
        self.color = color
        self.head = head

    def move(self, dir):
        self.pos[0] += dir[0]
        self.pos[0] %= MAP_WIDTH

        self.pos[1] += dir[1]
        self.pos[1] %= MAP_HEIGHT

    def update(self):
        pass

    def draw(self):
        drawing_pos = (self.pos[0] * RECT_WIDTH + RECT_WIDTH//2, self.pos[1] * RECT_HEIGHT + RECT_HEIGHT//2+1)
        ar.draw_rectangle_filled(*drawing_pos, RECT_WIDTH - GAP, RECT_HEIGHT - GAP, self.color)
        if self.head:
            ar.draw_rectangle_filled(*drawing_pos, (RECT_WIDTH-GAP)//2, (RECT_HEIGHT-GAP)//2, (np.array(self.color)*0.3).astype("int8"))
            ar.draw_rectangle_outline(*drawing_pos, (RECT_WIDTH-GAP)//2, (RECT_HEIGHT-GAP)//2, WHITE, 2)

    def lays_on(self, pos):
        return pos[0] == self.pos[0] and pos[1] == self.pos[1]


class Snake:
    def __init__(self, pos, game, brain, id):
        self.id = id
        self.game = game
        self.color = random.choice(COLORS)
        self.body = [BodySegment(self.color, pos, head=True)]
        self.history = deque([])
        self.brain = brain
        self.timer = max(MAP_HEIGHT,MAP_WIDTH)*TIMER_RATIO
        self.time_left = self.timer
        self.ticks = 0
        self.longest = 1

    def move(self):
        left = MAP_WIDTH
        right = MAP_WIDTH
        down = MAP_HEIGHT
        up = MAP_HEIGHT

        for snake_id in self.game.snakes:
            snake = self.game.snakes[snake_id]
            for seg in snake.body:
                diff_x = abs(seg.pos[0] - self.body[0].pos[0])
                diff_y = abs(seg.pos[1] - self.body[0].pos[1])
                if diff_x + diff_y == 0:
                    pass
                elif diff_y == 0:
                    if seg.pos[0] < self.body[0].pos[0]:
                        left = min(diff_x, left)
                        right = min(MAP_WIDTH - diff_x, right)
                    else:
                        right = min(diff_x, right)
                        left = min(MAP_WIDTH - diff_x, left)

                elif diff_x == 0:
                    if seg.pos[1] < self.body[0].pos[1]:
                        down = min(diff_y, down)
                        up = min(MAP_HEIGHT - diff_y, up)
                    else:
                        up = min(diff_y, up)
                        down = min(MAP_HEIGHT - diff_y, down)

        input_snakes = (np.array([MAP_HEIGHT, MAP_WIDTH, MAP_HEIGHT, MAP_WIDTH]) - np.array([up, right, down, left]))**10/np.array([MAP_HEIGHT-1, MAP_WIDTH-1, MAP_HEIGHT-1, MAP_WIDTH-1])**10

        left = MAP_WIDTH
        right = MAP_WIDTH
        down = MAP_HEIGHT
        up = MAP_HEIGHT
        for apple in self.game.apples:
            diff_x = abs(apple.pos[0] - self.body[0].pos[0])
            diff_y = abs(apple.pos[1] - self.body[0].pos[1])
            if diff_y == 0:
                if apple.pos[0] < self.body[0].pos[0]:
                    left = min(diff_x, left)
                    right = min(MAP_WIDTH - diff_x, right)
                else:
                    right = min(diff_x, right)
                    left = min(MAP_WIDTH - diff_x, left)

            elif diff_x == 0:
                if apple.pos[1] < self.body[0].pos[1]:
                    down = min(diff_y, down)
                    up = min(MAP_HEIGHT - diff_y, up)
                else:
                    up = min(diff_y, up)
                    down = min(MAP_HEIGHT - diff_y, down)

        input_apple = (np.array([MAP_HEIGHT, MAP_WIDTH, MAP_HEIGHT, MAP_WIDTH]) - np.array([up, right, down, left]))**2/np.array([MAP_HEIGHT-1, MAP_WIDTH-1, MAP_HEIGHT-1, MAP_WIDTH-1])**2
        dir = self.brain.decide(np.concatenate((input_snakes, input_apple)))

        for snake_id in self.game.snakes:
            if self.game.snakes[snake_id].lays_on([self.body[0].pos[0] + dir[0], self.body[0].pos[1] + dir[1]]):
                self.die()
                return

        self.history.appendleft(dir)
        last_pos = self.body[-1].pos.copy()

        for i in range(len(self.body)):
            self.body[i].move(self.history[i])
        for i in range(len(self.game.apples)):
            if self.game.apples[i].lays_on(self.body[0].pos):
                self.body.append(BodySegment(self.color, last_pos))
                self.game.apples[i].eat()
                if len(self.body)>self.longest:
                    self.longest = len(self.body)

        if len(self.history) >= len(self.body):
            self.history.pop()

        self.time_left -= 1
        self.ticks += 1
        if self.time_left < 1:
            self.time_left += self.timer
            del self.body[-1]
            if len(self.body) == 0:
                self.die()
                return

    def update(self):
        pass

    def draw(self):
        for segment in self.body:
            segment.draw()

    def lays_on(self, pos):
        for seg in self.body:
            if seg.lays_on(pos):
                return True
        return False

    def die(self):
        self.game.snake_died(self.id)


class Apple(BodySegment):
    def __init__(self, game):
        self.game = game
        super().__init__(RED, [0, 0])
        self.grow()

    def grow(self):
        self.pos = self.game.get_free_place()

    def eat(self):
        self.grow()



import arcade as ar
import numpy as np
import random


class Brain:
    def __init__(self):
        pass

    def decide(self, input_data):
        pass


class Dumb(Brain):
    def __init__(self, dir):
        self.dir = dir
        super().__init__()

    def decide(self, input_data):
        return self.dir


class User(Brain):
    def __init__(self, game):
        super().__init__()
        self.game = game
        self.last = [0, 1]

    def decide(self, input_data):
        if self.game.buttons[ar.key.UP]:
            self.last = [0, 1]
            return [0, 1]
        elif self.game.buttons[ar.key.RIGHT]:
            self.last = [1, 0]
            return [1, 0]
        elif self.game.buttons[ar.key.DOWN]:
            self.last = [0, -1]
            return [0, -1]
        elif self.game.buttons[ar.key.LEFT]:
            self.last = [-1, 0]
            return [-1, 0]
        else:
            return self.last


class NN(Brain):
    def __init__(self, game, W1=None, W2=None):
        super().__init__()
        self.game = game
        if W1 is None or W2 is None:
            self.W = (np.random.rand(4, 9) - 0.5) * 2
        else:
            net = np.random.rand(*W1.shape) > 0.5
            self.W = W1 * net + W2 * np.logical_not(net) - 1
            if random.random() > 0.6:
                self.mutate()

    def decide(self, input_data):
        answer = [(0, 1), (1, 0), (0, -1), (-1, 0)]
        x = np.concatenate((np.array([1]), input_data))
        y = x.dot(self.W.T)
        return answer[list(y).index(np.max(y))]

    def mutate(self):
        random_W = ((np.random.rand(4, 9)) * 3).astype("int8")-1
        net = np.random.rand(*self.W.shape) > 0.3
        self.W = net * self.W + np.logical_not(net) * (random_W*0.6 + self.W*0.4)
        negating_net = np.ones(self.W.shape) - (np.random.rand(*self.W.shape) > 0.95) * 2
        self.W = self.W * negating_net



from distutils.core import setup

setup(
    name="aisnakes",
    version="1.0",
    scripts=["aisnakes", "config.py", "brains.py", "snakes.py"],
    install_requires=["numpy", "arcade"],

    author= "JanKaifer",
    author_email="kaifer741@gmail.com",
    description="Snakes learning to live with one vs all perceptrons",
    keywords=["snake", "ai", "perceptron", "machine", "learning", "cool", "useless"],
    url="https://github.com/JanKaifer/AISnakes"
)



SCREEN_WIDTH = 1900
SCREEN_HEIGHT = 1000

RECT_WIDTH = 25
RECT_HEIGHT = 25

GAP = 5

APPLES = 15
SNAKES = 10
TIMER_RATIO = 0.4

GRID_COLOR = (10, 10, 10)
BACKGROUND_COLOR = (100, 200, 255)
SNAKES_COLOR = (255, 100, 0)

BLACK = (0, 0, 0)
RED = (255, 0, 0)
GREEN = (0, 255, 0)
BLUE = (0, 0, 255)
WHITE = (255, 255, 255)
LIGHT_GREEN = (136, 255, 127)
SKY_BLUE = (127, 218, 255)
YELLOW = (242, 255, 127)
PINK = (255, 127, 240)
BROWN = (132, 39, 0)
DARK_GREEN = (28, 56, 24)

COLORS = (GREEN, BLUE, LIGHT_GREEN, YELLOW, PINK, BROWN, DARK_GREEN)

MAP_HEIGHT = SCREEN_HEIGHT // RECT_HEIGHT
MAP_WIDTH = SCREEN_WIDTH // RECT_WIDTH
